from ._load import load_ontology, load_minimal_ontology
from ._factory import TermFactory, MinimalTermFactory

__all__ = [
    'load_ontology', 'load_minimal_ontology',
    'TermFactory', 'MinimalTermFactory'
]
