from hpotk.model import TermId

# Descendents of Severity `HP:0012824`.

SEVERITY: TermId = TermId.from_curie("HP:0012824")

BORDERLINE: TermId = TermId.from_curie("HP:0012827")
MILD: TermId = TermId.from_curie("HP:0012825")
MODERATE: TermId = TermId.from_curie("HP:0012826")
SEVERE: TermId = TermId.from_curie("HP:0012828")
PROFOUND: TermId = TermId.from_curie("HP:0012829")
