# TODO[v1.0.0] - remove the deprecated methods
from ._traversal import get_ancestors, get_parents
from ._traversal import get_children, get_descendants
from ._traversal import exists_path
from . import similarity
